#include "libasset.h"
#include <sys/stat.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <zlib.h>

    /* debugging flag */
int libasset_debug = 0;

    /* Defines an asset handle. */
struct AssetHandle
{
    struct AssetSource	*source;	// Asset source this handle is from
    FILE		*fh;		// Pointer to file handle
    long		size;		// Size of asset, uncompressed
    struct HashTableEntry *entry;	// Pointer to hash table entry, if any
    long		bytesRead;	// Number of uncomp. bytes read so far
    long		dataOffset;	// Data offset of compressed data in archive.
    char		*zbuffer;	// Decompression buffer
    z_stream		zinfo;		// Decompression state
};

    /* For zip files, we store the directory in a hash table. */
struct HashTableEntry
{
    struct HashTableEntry *nextInHash;	// Next asset in this hash chain
    struct HashTableEntry *nextInDir;	// Next asset in directory (if we want to iterate)
    long		compressedSize;	// Compressed size of asset
    long		uncompressedSize; // Uncompressed size of asset
    long		fileOffset;	// Offset of file header in .zip file
    long		compressionMethod; // Compression method
    unsigned long	hashVal;	// Hash value, for searching
	// Null-terminated name follows struct
};

    /* Asset Manager, it's just a linked list of asset sources. */
struct AssetManager
{
    struct AssetSource	*first;		// First asset source in list
    struct AssetSource	*last;		// Last asset source in list
    int			err_code;	// error code
    char		msg[ 256 ];	// error message
};

    /* A source of assets, either a zip file or a directory. */
struct AssetSource
{
    struct AssetManager	*mgr;
    struct AssetSource	*next;		// next asset source in path list
    char		*path;		// path to file or directory containing assets
    char		is_dir;		// TRUE if this is a directory; FALSE if zipfile
    FILE		*fh;		// File Handle
    struct HashTableEntry **hashTable;	// For zipfiles, here's a copy of the directory.
    struct HashTableEntry *first;	// First entry in directory
    struct HashTableEntry *last;	// last entry in directory
};

    /* Compression methods supported in zip files. */
enum ECompressionMethods {
    Compression_Stored = 0,
    Compression_Deflated = 8
};

#define HASH_SIZE	127		// Size of hash table
#define ZBUFFER_SIZE	4096		// Size of decompression buffer

    /* Various structs for decoding zip files. */
struct ECRec
{
    unsigned char	sig[ 4 ];
    unsigned char	thisDiskIndex[ 2 ];	// Number of this disk
    unsigned char	directoryDiskIndex[ 2 ];// Number of disk w/start of directory
    unsigned char	directoryEntries[ 2 ];	// Number of directory entries this disk
    unsigned char	totalDirectoryEntries[ 2 ];// Total entries in directory
    unsigned char	directorySize[ 4 ];	// Size of central directory
    unsigned char	directoryOffset[ 4 ];	// Offset of central directory
    unsigned char	commentLength[ 2 ];	// Length of zipfile comment
};

struct CDirEntry {
    unsigned char	versionMadeBy[ 2 ];
    unsigned char	versionNeededToExtract[ 2 ];
    unsigned char	generalPurposedBitFlag[ 2 ];
    unsigned char	compressionMethod[ 2 ];
    unsigned char	lastModFileTime[ 2 ];
    unsigned char	lastModFileDate[ 2 ];
    unsigned char	crc32[ 4 ];
    unsigned char	compressedSize[ 4 ];
    unsigned char	uncompressedSize[ 4 ];
    unsigned char 	filenameLength[ 2 ];
    unsigned char	extraFieldLength[ 2 ];
    unsigned char	fileCommentLength[ 2 ];
    unsigned char	diskNumberStart[ 2 ];
    unsigned char	internalFileAttributes[ 2 ];
    unsigned char	externalFileAttributes[ 4 ];
    unsigned char	relativeOffsetLocalHeader[ 4 ];
};

struct LDirEntry {
    unsigned char	versionNeededToExtract[ 2 ];
    unsigned char	generalPurposedBitFlag[ 2 ];
    unsigned char	compressionMethod[ 2 ];
    unsigned char	lastModFileTime[ 2 ];
    unsigned char	lastModFileDate[ 2 ];
    unsigned char	crc32[ 4 ];
    unsigned char	compressedSize[ 4 ];
    unsigned char	uncompressedSize[ 4 ];
    unsigned char 	filenameLength[ 2 ];
    unsigned char	extraFieldLength[ 2 ];
};

    /* Calculate a hash value for a string. This is the standard hash
       algorithm I always use, stolen from the AmigaDOS file system.
    */
static unsigned long hash( char *inName, int length )
{
    unsigned long	h = 0;

    while (length--) h = h * 13 + *inName++;
    return h;
}

    /* Endian-swap a 16-bit integer. */
static int makeword( unsigned char *ptr )
{
    return ptr[ 0 ] | (ptr [ 1 ] << 8);
}

    /* Endian-swap a 32-bit integer. */
static int makelong( unsigned char *ptr )
{
    return ptr[ 0 ] | (ptr [ 1 ] << 8) | (ptr [ 2 ] << 16) | (ptr [ 3 ] << 24);
}

    /** Open a handle to an asset. */
ASSET_HANDLE open_asset( ASSET_MANAGER mgr, char *name )
{
    struct AssetSource	*src;
    ASSET_HANDLE	handle;
    char		*save_path = NULL;

    if (mgr == NULL) return NULL;
    if (name == NULL)
    {
	mgr->err_code = EINVAL;
	strcpy( mgr->msg, "NULL asset name." );
	return NULL;
    }

	// Search each source for the item.
    for (src = mgr->first; src != NULL; src = src->next )
    {
	if (src->is_dir)	// If it's a directory
	{
		// Open the file from the directory
	    struct stat		st;
	    FILE		*fh;

		// Change directory to that directory, and then
		// open the file, if it exists
	    if (save_path == NULL) save_path = alloca( MAXPATHLEN );
	    getcwd( save_path, MAXPATHLEN );
	    chdir( src->path );
	    fh = fopen( name, "rb" );
	    if (fh == NULL)
	    {
		int	err = errno;

		chdir( save_path );

		    // If the file didn't exist, then try one of the other
		    // asset sources; Otherwise, report an error.
		if (err == ENOENT) continue;
		mgr->err_code = err;
		snprintf( mgr->msg, sizeof mgr->msg, "%s: %s\n", strerror( err ), name );
		return NULL;
	    }
	    chdir( save_path );

		// Stat the file.
	    if (fstat( fileno( fh ), &st ) != 0)
	    {
		mgr->err_code = errno;
		snprintf( mgr->msg, sizeof mgr->msg, "%s: %s\n", strerror( errno ), name );
		return NULL;
	    }

		// Allocate an asset handle and fill it in.
	    if ((handle = malloc( sizeof *handle )) == NULL)
	    {
		mgr->err_code = ENOMEM;
		strcpy( mgr->msg, "Insufficient memory" );
		return NULL;
	    }

	    handle->source = src;
	    handle->entry = NULL;
	    handle->size = st.st_size;
	    handle->bytesRead = 0;
	    handle->dataOffset = 0;
	    handle->fh = fh;

	    return handle;
	}
	else if (src->hashTable == NULL)
	{
		// If there's no entries in this zip file, then
		// ignore it.
	    continue;
	}
	else
	{
	    unsigned long	hashVal = hash( name, strlen( name ) );
	    int			hashIndex = hashVal % HASH_SIZE;
	    struct HashTableEntry *ht;

		// Search the hash table
	    for (ht = src->hashTable[ hashIndex ];
		 ht != NULL;
		 ht = ht->nextInHash)
	    {
		if (ht->hashVal == hashVal
		    && strcmp( name, (char *)(ht + 1) ) == 0)
		{
		    struct LDirEntry	*lrec;
			    
			// We only support the standard compression
			// methods.
		    if (ht->compressionMethod != Compression_Stored
			&& ht->compressionMethod != Compression_Deflated)
		    {
			mgr->err_code = EINVAL;
			snprintf( mgr->msg,
				  sizeof mgr->msg,
				  "Unsupported compression method: ",
				  name );
			return NULL;
		    }

			// Allocate space for the local file header.
		    if ((lrec = alloca( sizeof (struct LDirEntry) )) == NULL)
		    {
			mgr->err_code = ENOMEM;
			strcpy( mgr->msg, "Insufficient memory" );
			return NULL;
		    }

			// Read in the local header...
		    if (fseek( src->fh, ht->fileOffset + 4, SEEK_SET ) != 0
			|| fread( lrec, 1, sizeof *lrec, src->fh ) != sizeof *lrec)
		    {
			mgr->err_code = errno;
			snprintf( mgr->msg,
				  sizeof mgr->msg,
				  "%s: %s\n",
				  strerror( errno ),
				  name );
			return NULL;
		    }

			// Allocate a handle for this resource.
			// Q: We might want to dup the file handle?
		    if ((handle = malloc( sizeof *handle )) == NULL)
		    {
			mgr->err_code = ENOMEM;
			strcpy( mgr->msg, "Insufficient memory" );
			return NULL;
		    }

		    handle->source = src;
		    handle->entry = ht;
		    handle->size = ht->uncompressedSize;
		    handle->bytesRead = 0;
		    handle->fh = src->fh;
		    handle->dataOffset = ht->fileOffset + 4
			+ sizeof *lrec
			+ makeword( lrec->filenameLength )
			+ makeword( lrec->extraFieldLength );

		    if (libasset_debug)
		    {
			printf( "Opening handle for %s, cMethod=%d, data start = %d\n",
				name,
				ht->compressionMethod,
				handle->dataOffset );
		    }

		    if (ht->compressionMethod == Compression_Deflated)
		    {
			int		status;
			int		initial_in = ht->compressedSize;
			
			if (initial_in > ZBUFFER_SIZE)
			    initial_in = ZBUFFER_SIZE;
			    
			if ((handle->zbuffer = malloc( ZBUFFER_SIZE )) == NULL)
			{
			    free( handle );
			    mgr->err_code = ENOMEM;
			    strcpy( mgr->msg, "Insufficient memory" );
			    return NULL;
			}

			handle->zinfo.next_in = handle->zbuffer;
			handle->zinfo.avail_in = initial_in;

			    // Read in the first block of compressed data
			if (fseek( handle->fh, handle->dataOffset, SEEK_SET ) != 0
			    || fread( handle->zbuffer, 1, initial_in, handle->fh) != initial_in)
			{
			    mgr->err_code = errno;
			    snprintf( mgr->msg,
				      sizeof mgr->msg, "%s: %s\n",
				      strerror( errno ),
				      handle->entry + 1 );
			    free( handle->zbuffer );
			    free( handle );
			    return NULL;
			}

			handle->zinfo.opaque = NULL;
			handle->zinfo.zalloc = Z_NULL;
			handle->zinfo.zfree = Z_NULL;

			    // Initialize zlib with undocumented "nowrap" option.
			status = inflateInit2( &handle->zinfo, -15 );

			if (libasset_debug)
			{
			    printf( "inflateInit2 returned %d, avail_in = %d, total_in = %d\n",
				    status, handle->zinfo.avail_in, handle->zinfo.total_in );
			}

			if (status != Z_OK)
			{
			    if (status == Z_MEM_ERROR)
			    {
				mgr->err_code = ENOMEM;
				strcpy( mgr->msg, "Insufficient memory" );
			    }
			    else if (status == Z_VERSION_ERROR)
			    {
				mgr->err_code = EINVAL;
				strcpy( mgr->msg, "Unsupported compression version" );
			    }
			    else
			    {
				mgr->err_code = EINVAL;
				strcpy( mgr->msg, "Unknown error" );
			    }

			    free( handle );
			    return NULL;
			}
		    }
		    return handle;
		}
	    }
	}
    }

    mgr->err_code = ENOENT;
    snprintf( mgr->msg, sizeof mgr->msg, "Not found: %s\n", name );
    return NULL;
}

int close_asset( ASSET_HANDLE handle )
{
    if (handle == NULL) return -1;

    if (handle->entry == NULL)
    {
	fclose( handle->fh );
    }
    else if (handle->entry->compressionMethod == Compression_Deflated)
    {
	    // Free compression resources
	inflateEnd( &handle->zinfo );
    }

    free( handle );
    return 0;
}

long read_asset( ASSET_HANDLE handle, char *buffer, long length )
{
    long		result;
    struct AssetManager	*mgr;

    if (handle == NULL) return -1;

    mgr = handle->source->mgr;

    if (length < 0)
    {
	mgr->err_code = EINVAL;
	strcpy( mgr->msg, "Invalid length parameter to read_asset" );
	return -1;
    }

    if (handle->entry == NULL)
    {
	    // A single-file asset.
	    // Don't read past end of file
	if (length > handle->size - handle->bytesRead)
	    length = handle->size - handle->bytesRead;
	if (length == 0) return 0;
	
	result = fread( buffer, 1, length, handle->fh );
	if (result < 0)
	{
	    mgr->err_code = errno;
	    snprintf( mgr->msg, sizeof mgr->msg, "%s\n", strerror( errno ) );
	    return -1;
	}
	handle->bytesRead += result;
	return result;
    }
    else if (handle->entry->compressionMethod == Compression_Stored)
    {
	    // An uncompressed asset in an archive file
	    // Don't read past end of asset.
	if (length > handle->size - handle->bytesRead)
	    length = handle->size - handle->bytesRead;
	if (length == 0) return 0;

	if (fseek( handle->fh, handle->dataOffset + handle->bytesRead, SEEK_SET ) != 0
	    || fread( buffer, 1, length, handle->fh ) != length)
	{
	    mgr->err_code = errno;
	    snprintf( mgr->msg,
		      sizeof mgr->msg,
		      "%s: %s\n",
		      strerror( errno ),
		      handle->entry + 1 );
	    return -1;
	}
	handle->bytesRead += length;
	return length;
    }
    else
    {
	    // A compressed asset in an archive file.

	    // The output buffer is the buffer passed in to read_asset,
	    // where the application wants us to place the data. The input
	    // buffer is the private internal buffer where we store the
	    // compressed data read from the zip file.

	    // Set up the output buffer
	handle->zinfo.next_out = buffer;
	handle->zinfo.avail_out = length;

	for (;;)
	{
		// If the output buffer is full, then return
	    if (handle->zinfo.avail_out <= 0)
	    {
		handle->bytesRead += length;
		return length;
	    }

		// If there's no more data in the input buffer
	    if (handle->zinfo.avail_in == 0)
	    {
		long	clength;

		    // Calculate how much we're going to read.
		clength = handle->size - handle->zinfo.total_in;
		if (clength > ZBUFFER_SIZE) clength = ZBUFFER_SIZE;

		    // If there's in fact anything at all left to read.
		if (clength > 0)
		{
			// Seek to the correct position in the file
		    result = fseek( handle->fh,
				    handle->dataOffset + handle->zinfo.total_in,
				    SEEK_SET );
		    
			// Read clength worth of bytes
		    result = fread( handle->zbuffer, 1, clength, handle->fh );
		    if (result < 0)
		    {
			mgr->err_code = errno;
			snprintf( mgr->msg,
				  sizeof mgr->msg, "%s: %s\n",
				  strerror( errno ),
				  handle->entry + 1 );
			return -1;
		    }

		    if (libasset_debug)
		    {
			printf( "read %d bytes of compressed data\n", result );
		    }
		    
		    handle->zinfo.avail_in = result;
		    handle->zinfo.next_in = handle->zbuffer;
		}
	    }

	    result = inflate( &handle->zinfo, Z_NO_FLUSH );
	    if (libasset_debug)
	    {
		printf( "inflate returned = %d, avail_in = %d, total_in = %d\n",
			result, handle->zinfo.avail_in, handle->zinfo.total_in );
	    }
	    
	    if (result == Z_STREAM_END)
	    {
		result = length - handle->zinfo.avail_out;
		
		handle->bytesRead += result;

		if (handle->bytesRead != handle->size)
		{
		    mgr->err_code = -1;
		    snprintf( mgr->msg,
			      sizeof mgr->msg, "Compression size mismatch: %s\n",
			      handle->entry + 1 );
		    return -1;
		}
		
		return result;
	    }
	    else if (result == Z_OK)
	    {
		if (handle->zinfo.avail_in == 0
		    && handle->zinfo.total_in >= handle->size)
		{
			// No more input data...
		    result = length - handle->zinfo.avail_out;
		    handle->bytesRead += result;
		    return result;
		}
	    }
	    else
	    {
		char	*zerr = "UNKNOWN";
		switch (result) {
		case Z_NEED_DICT:	zerr = "Z_NEED_DICT"; break;
		case Z_ERRNO:		zerr = "Z_ERRNO"; break;
		case Z_STREAM_ERROR:	zerr = "Z_STREAM_ERROR"; break;
		case Z_DATA_ERROR:	zerr = "Z_DATA_ERROR"; break;
		case Z_MEM_ERROR:	zerr = "Z_MEM_ERROR"; break;
		case Z_BUF_ERROR:	zerr = "Z_BUF_ERROR"; break;
		case Z_VERSION_ERROR:	zerr = "Z_VERSION_ERROR"; break;
		}
		
//		printf( "got error.\n" );
		snprintf( mgr->msg,
			  sizeof mgr->msg,
			  "zlib error: %s: %s\n",
			  zerr,
			  handle->entry + 1 );
		    
		return -1;
	    }
	}
    }
}

int is_end_of_asset( ASSET_HANDLE handle )
{
    return handle->bytesRead >= handle->size;
}

long asset_size( ASSET_HANDLE handle )
{
    return handle->size;
}

long asset_pos( ASSET_HANDLE handle )
{
    return handle->bytesRead;
}

char *get_asset_error( ASSET_MANAGER mgr, int *errCode )
{
    if (errCode) *errCode = mgr->err_code;
    return mgr->msg;
}

static struct AssetSource *open_asset_source( ASSET_MANAGER mgr, char *path )
{
    struct stat		st;
    int			err;
    struct AssetSource	*src;

    if (libasset_debug)
    {
	printf( "Opening asset.\n" );
    }

    if (mgr == NULL || path == NULL) return NULL;

    mgr->err_code = 0;

	/* Stat the file */
    if (stat( path, &st ) != 0)
    {
	mgr->err_code = errno;
	snprintf( mgr->msg, sizeof mgr->msg, "%s: %s\n", strerror( errno ), path );
	return NULL;
    }

    if (S_ISREG( st.st_mode ))		/* It's a regular file. */
    {
	char	sig[ 4 ];
	char	*buffer;
	char	*dirBuffer = NULL;
	int	tail_size;
	char	*scan;
	int	sig_offset;
	struct ECRec *ecrec;
	int	directoryCount;
	int	directorySize;
	int	directoryOffset;
	int	i;
	int	hashIndex;

	if ((src = malloc( sizeof *src )) == NULL) goto handle_nomem;

	bzero( src, sizeof *src );
	src->path = path;
	src->fh = fopen( src->path, "rb" );
	src->mgr = mgr;

	if (src->fh == NULL) goto handle_errno;
	tail_size = st.st_size;
	if (tail_size > 1024) tail_size = 1024;

	    // Extract the asset from the zip file.
	buffer = alloca( tail_size );

	if (libasset_debug)
	{
	    printf( "Beginning search for ecrec.\n" );
	}
	
	if (fseek( src->fh, st.st_size - tail_size, SEEK_SET ) != 0)
	    goto handle_errno;

	if (fread( buffer, 1, tail_size, src->fh ) != tail_size)
	    goto handle_errno;

	for (scan = &buffer[ tail_size - 4 ]; scan >= buffer; scan--)
	{
		// Note: The reason we're not doing a memcmp
		// here is that storing the signature in an array
		// might cause this executable image (libasset)
		// to be mis-identified as a zip file.
	    if (scan[ 0 ] == 'P'
		&& scan[ 1 ] == 'K'
		&& scan[ 2 ] == 5
		&& scan[ 3 ] == 6)
	    {
		break;
	    }
	}

	    // If we didn't find the signature.
	if (scan < buffer)
	{
	    mgr->err_code = EINVAL;
	    snprintf( mgr->msg, sizeof mgr->msg, "Not a zip file: %s", path );
	    goto cleanup;
	}

	sig_offset = scan - buffer + st.st_size - tail_size;

	if (st.st_size - sig_offset < sizeof (struct ECRec))
	{
	    mgr->err_code = EINVAL;
	    snprintf( mgr->msg, sizeof mgr->msg, "Not a valid zip file: %s", path );
	    goto cleanup;
	}

	ecrec = (struct ECRec *)scan;

	directoryCount = makeword( ecrec->directoryEntries );
	directorySize = makelong( ecrec->directorySize );
	directoryOffset = makelong( ecrec->directoryOffset );

	if (makeword( ecrec->thisDiskIndex ) != 0
	    || makeword( ecrec->directoryDiskIndex ) != 0
	    || makeword( ecrec->totalDirectoryEntries ) != directoryCount)
	{
	    mgr->err_code = EINVAL;
	    snprintf( mgr->msg, sizeof mgr->msg, "Multi-disk archives not supported: %s", path );
	    goto cleanup;
	}

	if (libasset_debug)
	{
	    printf( "directory for %s: count = %d, size = %d, offset = %x, ecrec_expect = %x\n",
		    path,
		    directoryCount,
		    directorySize,
		    directoryOffset,
		    directoryOffset + directorySize);
	}
	

	dirBuffer = alloca( directorySize );

	if (fseek( src->fh, directoryOffset, SEEK_SET ) != 0)
	    goto handle_errno;

	if (fread( dirBuffer, 1, directorySize, src->fh ) != directorySize)
	    goto handle_errno;

	if (dirBuffer[ 0 ] != 'P'
	    || dirBuffer[ 1 ] != 'K'
	    || dirBuffer[ 2 ] != 1
	    || dirBuffer[ 3 ] != 2)
	{
	    mgr->err_code = EINVAL;
	    snprintf( mgr->msg, sizeof mgr->msg, "Zip file directory not found: %s", path );
	    goto cleanup;
	}

	scan = dirBuffer;
	for (i = 0; i < directoryCount; i++)
	{
	    struct CDirEntry	*dirEntry;
	    long		compressedSize;
	    long		uncompressedSize;
	    short		filenameLength;
	    short		extraFieldLength;
	    short		fileCommentLength;
	    short		compressionMethod;
	    long		relativeOffset;
//	    long		headerLength;
	    char		*name;
	    struct HashTableEntry *ht;

	    if (scan[ 0 ] != 'P'
		|| scan[ 1 ] != 'K'
		|| scan[ 2 ] != 1
		|| scan[ 3 ] != 2)
	    {
		mgr->err_code = EINVAL;
		snprintf( mgr->msg, sizeof mgr->msg, "Invalid zip file entry: %s", path );
		goto cleanup;
	    }

	    scan += 4;
	    dirEntry = (struct CDirEntry *)scan;

	    compressedSize = makelong( dirEntry->compressedSize );
	    uncompressedSize = makelong( dirEntry->uncompressedSize );
	    filenameLength = makeword( dirEntry->filenameLength );
	    extraFieldLength = makeword( dirEntry->extraFieldLength );
	    fileCommentLength = makeword( dirEntry->fileCommentLength );
	    compressionMethod = makeword( dirEntry->compressionMethod );
	    relativeOffset = makelong( dirEntry->relativeOffsetLocalHeader );

	    scan += sizeof *dirEntry;

	    if (src->hashTable == NULL)
	    {
		src->hashTable = calloc( HASH_SIZE, sizeof (struct HashTableEntry *) );
		if (src->hashTable == NULL) goto handle_nomem;
	    }

	    if ((ht = malloc( filenameLength + 1 + sizeof (struct HashTableEntry) )) == NULL)
		goto handle_nomem;

	    name = (char *)(ht + 1);
	    name[ filenameLength ] = '\0';
	    memcpy( name, scan, filenameLength );

	    if (libasset_debug)
	    {
		printf( "%d ==> %d: %s [offset %d/$%x]\n",
			uncompressedSize,
			compressedSize,
			name,
			relativeOffset,
			relativeOffset );
	    }

	    scan += filenameLength + extraFieldLength + fileCommentLength;

	    ht->compressedSize = compressedSize;
	    ht->uncompressedSize = uncompressedSize;
	    ht->fileOffset = relativeOffset;
	    ht->compressionMethod = compressionMethod;
	    ht->hashVal = hash( name, filenameLength );

		// Link into hash table
	    hashIndex = ht->hashVal % HASH_SIZE;
	    ht->nextInHash = src->hashTable[ hashIndex ];
	    src->hashTable[ hashIndex ] = ht;

		// Create singly-linked list of ht entries.
	    ht->nextInDir = NULL;
	    if (src->last != NULL) src->last->nextInDir = ht;
	    else src->first = ht;
	    src->last = ht;
	}

	return src;
    }
    else if (S_ISDIR( st.st_mode ))	/* It's a directory. */
    {
	if ((src = malloc( sizeof *src )) == NULL)
	{
	    mgr->err_code = ENOMEM;
	    strcpy( mgr->msg, "Insufficient memory" );
	    return NULL;
	}

	src->next = NULL;
	src->path = path;
	src->is_dir = 1;
	src->hashTable = NULL;
	src->fh = NULL;

	return src;
    }
    else return NULL;

 handle_nomem:
    mgr->err_code = ENOMEM;
    strcpy( mgr->msg, "Insufficient memory" );
    goto cleanup;

 handle_errno:
    mgr->err_code = errno;
    snprintf( mgr->msg, sizeof mgr->msg, "%s: %s\n", strerror( errno ), path );

 cleanup:
    if (src->fh != NULL) fclose( src->fh );
    free( src );
    return NULL;
}

static struct AssetSource *close_asset_source( struct AssetSource *src )
{
	// Free up all structures used by asset source.
    while (src->first != NULL)
    {
	struct HashTableEntry	*next = src->first->nextInDir;

	free( src->first );
	src->first = next;
    }
    if (src->hashTable) free( src->hashTable );
    if (src->fh) fclose( src->fh );
    if (src->path) free( src->path );

    free( src );
}

ASSET_MANAGER new_asset_manager()
{
    ASSET_MANAGER	mgr;

    if ((mgr = malloc( sizeof *mgr )) == NULL)
	return NULL;

    bzero( mgr, sizeof *mgr );

    return mgr;
}

void delete_asset_manager( ASSET_MANAGER mgr )
{
	// Delete all the asset sources.
    while (mgr->first != NULL)
    {
	struct AssetSource	*src = mgr->first;

	mgr->first = src->next;
	close_asset_source( src );
    }
}

int set_asset_path( ASSET_MANAGER mgr, char *path )
{
    char	*start = path;
    char	*scan = path;
    int		result = 0;

    if (mgr == NULL || path == NULL) return -1;

	// Remove all existing paths.
    while (mgr->first != NULL)
    {
	struct AssetSource	*src = mgr->first;

	mgr->first = src->next;
	close_asset_source( src );
    }
    mgr->first = mgr->last = NULL;

	// Scan through path string
    for (;;)
    {
	if (*scan == '\0' || *scan == ':')
	{
	    if (scan - start > 0)
	    {
		struct AssetSource	*src;
		char			*str;

		    // Make a copy of the string fragment.
		str = malloc( scan - start + 1 );
		if (str == NULL)
		{
		    mgr->err_code = ENOMEM;
		    strcpy( mgr->msg, "Insufficient memory" );
		    return;
		}
		bcopy( start, str, scan - start );
		str[ scan - start ] = '\0';

		    // Open the asset source
		src = open_asset_source( mgr, str );

		    // Link it into the list of sources
		if (src != NULL)
		{
		    if (mgr->last) mgr->last->next = src;
		    else mgr->first = src;

		    mgr->last = src;
		}
		else
		{
		    result = -1;
		    free( str );
		}
	    }
	    start = scan + 1;

	    if (*scan == '\0') break;
	}
	scan++;
    }
    return result;
}
