/* libasset.h -- interface of the 'libasset' asset management library
   version 0.1 June 26, 1999

    Copyright (C) 1999 Talin (talin@acm.org)
              
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#ifndef LIBASSETH
#define LIBASSETH

#ifdef __cplusplus
extern "C" {
#endif

/** @package libasset
    @author Talin
*/

/* libasset is a library intended to allow games or other applications
   to manage their runtime assets. The term "asset" is often used in
   the game industry to mean any file or fragment of data which has
   been authored or created by a human, such as a sound, piece of
   artwork, or a script. Often, it is convenient to store all of the
   assets in a "resource file" rather than in individual files - this
   makes installation and downloading simpler, and avoids some of the
   filesystem overhead (and end-user confusion) of having thousands
   of little files in the application's directory.

   Rather than defining a custom resource format, the libasset
   resource files are simply zip files created by the zip utility. The
   operation of the library has been intentionally modeled after the
   "classfiles" concept used in Java, whereby a "class path", which
   can include both zip files as well as ordinary directories, can be
   specified to the asset manager. The library can load ordinary files
   as well as compressed files from a zip archive.

   The advantage of the asset path strategy is that it allows easy
   replacement of a small number of assets without rebuilding the main
   asset file. This is especially useful in development, where an
   artist may want to override a number of assets in the resource file
   without having to constantly rebuild the resource file to
   incorporate the new artwork. In this case, all that is needed is to
   insert the artist's working directory into the asset path, before
   the main resource file.

   While libasset does cache the directory information of the zip
   files, it does not attempt to cache the actual assets
   themselves. This is generally better done at a higher level - for
   example, if an asset represents an image, then it's better to cache
   a pointer to the loaded-in image than the raw bytes which represent
   the archived form of the image.

   The library provides a minimal set of functions for reading asset
   data, which operate in a manner similar to the standard i/o
   functions.

   libasset is thread safe, provided the threads are using different
   instances of the structures. Each thread should have its own
   asset manager, and thus it's own asset handles. libasset does not
   protect itself against two threads using the same instance of a
   structure.
*/

typedef struct AssetHandle	*ASSET_HANDLE;
typedef struct AssetManager	*ASSET_MANAGER;

    /** Construct a new asset manager. This must be done before any
	other libasset functions can be invoked.
    */
ASSET_MANAGER new_asset_manager();

    /** Destroy the asset manager. This deallocates all of the memory
	used by the manager (except asset handles, which you must close
	individually, before the manager is closed.)

	If this function fails for some reason, NULL is returned.
    */
void delete_asset_manager( ASSET_MANAGER mgr );

    /** Set the path for searching for assets. The path consists of a
	list of either directories or zip files, seperated by colons.
	Note that zip files with large comments are not supported.

	The function will return -1 if there was an error detected
	in one or more elements of the path. Note, however, that
	this may not prevent other elements of the path from
	functioning correctly. It returns 0 if all elements of the
	path initialized successfully.

	If you call this function a second time, the new path will
	replace the old one. This will also cause all of the
	directories of the zip files in the path to be re-scanned.
    */
int set_asset_path( ASSET_MANAGER mgr, char *path );

    /** If an asset function returns with an error result,
	you can get a text description of the error with this
	function. The optional errCode field can be used to
	return the numeric error code, usually derived from
	errno.
     */
char *get_asset_error( ASSET_MANAGER mgr, int *errCode );

    /** Open a handle to an asset. This is analogous to a file handle,
	and can be used to read data from the asset.
    */
ASSET_HANDLE open_asset( ASSET_MANAGER mgr, char *name );

    /** Close an asset handle. */
int close_asset( ASSET_HANDLE );

    /** Read data from an asset handle. Returns the number of bytes
	read if there was no error; 0 if end of file; and less than
	zero if there was an error.
     */
long read_asset( ASSET_HANDLE, char *buffer, long length );

    /** Test for end of asset. Returns non-zero if we have read up
	to the end of the asset.
     */
int is_end_of_asset( ASSET_HANDLE );

    /** Return the uncompressed size, in bytes, of this asset.
     */
long asset_size( ASSET_HANDLE );

    /** Return how many bytes have been read in the asset so far.
     */
long asset_pos( ASSET_HANDLE );

#ifdef __cplusplus
}
#endif
#endif
